import sys
import asyncio
import random
import gramscript.aio
from gramscript.aio.routing import (by_content_type, make_content_type_routing_table,
                                    lower_key, by_chat_command, make_routing_table,
                                    by_regex)


def random_key(msg):
    return random.choice([
        0,
        (1,),
        (2, ('a1',)),
        (3, ('a1', 'a2'), {'b1': 'b'}),
        (4, (), {'kw4': 4444, 'kw5': 'xyz'}),
        ((None,), ()),
    ])


def zero(msg):
    print('Zero')


def one(msg):
    print('One')


def two(msg, a1):
    print('Two', a1)


def three(msg, a1, a2, b1):
    print('Three', a1, a2, b1)


def none_tuple(msg):
    print('None tuple')


def none_of_above(msg, *args, **kwargs):
    print('None of above', msg, args, kwargs)


top_router = gramscript.aio.helper.Router(random_key, {0: zero,
                                                       1: one,
                                                       2: two,
                                                       3: three,
                                                       (None,): none_tuple,
                                                       None: none_of_above})


async def fake0():
    for _ in range(20):
        await top_router.route({})
    print()


class ContentTypeHandler(object):
    def on_text(self, msg, text):
        print('Text', msg, text)

    def on_photo(self, msg, photo):
        print('Photo', msg, photo)


def make_message_like(mm):
    for d in mm:
        d.update({'chat': {'type': 'private', 'id': 1000}})


async def fake1():
    top_router.key_function = by_content_type()
    top_router.routing_table = make_content_type_routing_table(
        ContentTypeHandler())

    del top_router.routing_table['video']
    top_router.routing_table[None] = none_of_above
    messages = [{'text': 'abc'}, {
        'photo': 'some photo'}, {'video': 'some video'}]
    make_message_like(messages)
    for _ in range(10):
        await top_router.route(random.choice(messages))
    print()


class CommandHandler(object):
    def on_start(self, msg):
        print('Command: start', msg)

    def on_settings(self, msg):
        print('Command: settings', msg)

    def on_invalid_text(self, msg):
        print('Invalid text', msg)

    def on_invalid_command(self, msg):
        print('Invalid command', msg)


command_handler = CommandHandler()
command_router = gramscript.aio.helper.Router(lower_key(by_chat_command()),
                                              make_routing_table(command_handler, [
                                                  'start',
                                                  'settings',
                                                  ((None,), command_handler.on_invalid_text),
                                                  (None, command_handler.on_invalid_command),
                                              ]))


async def fake2():
    top_router.routing_table['text'] = command_router.route
    messages = [{'text': '/start'}, {'text': '/SETTINGS'}, {'text': '/bad'},
                {'text': 'plain text'}, {'photo': 'some photo'}, {'video': 'some video'}]

    make_message_like(messages)
    for _ in range(20):
        await top_router.route(random.choice(messages))
    print()


class RegexHandler(object):
    def on_CS101(self, msg, match):
        print('Someone mentioned CS101 !!!', msg, match.groups())

    def on_CS202(self, msg, match):
        print('Someone mentioned CS202 !!!', msg, match.groups())

    def no_cs_courses_mentioned(self, msg):
        print('No CS courses mentioned ...', msg)

    def course_not_exist(self, msg, match):
        print(f'{match.group(1)} does not exist', msg)


regex_handler = RegexHandler()
regex_router = gramscript.aio.helper.Router(by_regex(lambda msg: msg['text'], '(CS[0-9]{3})'),
                                            make_routing_table(regex_handler, [
                                                'CS101',
                                                'CS202',
                                                ((None,), regex_handler.no_cs_courses_mentioned),
                                                (None, regex_handler.course_not_exist),
                                            ]))


async def fake3():
    command_router.routing_table[(None, )] = regex_router.route
    messages = [{'text': '/start'}, {'text': '/SETTINGS'}, {'text': '/bad'}, {'text': 'plain text'}, {'text': 'I want to take CS101.'},
                {'text': "I\'d rather take CS202."}, {'text': "Why don\'t you take CS303?"}, {'text': 'I hate computer science!'}, {'photo': 'some photo'}, {'video': 'some video'}]

    make_message_like(messages)
    for _ in range(30):
        await top_router.route(random.choice(messages))
    print()


TOKEN = sys.argv[1]

bot = gramscript.aio.Bot(TOKEN)
bot._router.routing_table['chat'] = top_router.route

loop = asyncio.get_event_loop()
loop.run_until_complete(fake0())
loop.run_until_complete(fake1())
loop.run_until_complete(fake2())
loop.run_until_complete(fake3())

print('Send me some messages ...')
loop.create_task(bot.message_loop())
loop.run_forever()



def _apply_entities(text, entities, escape_map, format_map):
    def inside_entities(i):
        return any(map(lambda e:
                       e['offset'] <= i < e['offset']+e['length'],
                       entities))

    # Split string into char sequence and escape in-place to
    # preserve index positions.
    seq = list(map(lambda c, i: escape_map[c] if c in escape_map and not inside_entities(
        i) else c, list(text), range(len(text))))  # escape special characters

    # Ensure smaller offsets come first
    sorted_entities = sorted(entities, key=lambda e: e['offset'])
    offset = 0
    result = ''

    for e in sorted_entities:
        f, n, t = e['offset'], e['length'], e['type']

        result += ''.join(seq[offset:f])

        if t in format_map:
            # apply format
            result += format_map[t](''.join(seq[f:f+n]), e)
        else:
            result += ''.join(seq[f:f+n])

        offset = f + n

    result += ''.join(seq[offset:])
    return result


def apply_entities_as_markdown(text, entities):
    """
    Format text as Markdown. Also take care of escaping special characters.
    Returned value can be passed to :meth:`.Bot.sendMessage` with appropriate
    ``parse_mode``.

    :param text:
        plain text

    :param entities:
        a list of `MessageEntity <https://core.telegram.org/bots/api#messageentity>`_ objects
    """
    escapes = {'*': '\\*', '_': '\\_', '[': '\\[', '`': '\\`'}
    formatters = {'bold': lambda s, e: f'*{s}*', 'italic': lambda s, e: f'_{s}_', 'text_link': lambda s,
                  e: f'[{s}](' + e['url'] + ')', 'text_mention': lambda s, e: f'[{s}](tg://user?id=' + str(e['user']['id']) + ')', 'code': lambda s, e: f'`{s}`', 'pre': lambda s, e: '```text\n' + s + '```'}

    return _apply_entities(text, entities, escapes, formatters)


def apply_entities_as_html(text, entities):
    """
    Format text as HTML. Also take care of escaping special characters.
    Returned value can be passed to :meth:`.Bot.sendMessage` with appropriate
    ``parse_mode``.

    :param text:
        plain text

    :param entities:
        a list of `MessageEntity <https://core.telegram.org/bots/api#messageentity>`_ objects
    """
    escapes = {'<': '&lt;', '>': '&gt;', '&': '&amp;'}
    formatters = {'bold': lambda s, e: f'<b>{s}</b>', 'italic': lambda s, e: f'<i>{s}</i>', 'text_link': lambda s, e: '<a href="' +
                  e['url'] + '">' + s + '</a>', 'text_mention': lambda s, e: '<a href="tg://user?id=' + str(e['user']['id']) + '">' + s + '</a>', 'code': lambda s, e: f'<code>{s}</code>', 'pre': lambda s, e: f'<pre>{s}</pre>'}

    return _apply_entities(text, entities, escapes, formatters)



import sys
import time
from pprint import pprint
import gramscript
from gramscript.namedtuple import (
    LabeledPrice, Invoice, PreCheckoutQuery, ShippingQuery, ShippingOption,
    SuccessfulPayment)
from gramscript.loop import MessageLoop

"""
This script tests the payment process:
1. Send an invoice
2. Receive a shipping query, respond with answerShippingQuery()
3. Receive a pre-checkout query, respond with answerPreCheckoutQuery()
4. Receive a successful payment

Run it by:
$ python2.7 script.py <bot-token> <payment-provider-token>
"""


def on_chat_message(msg):
    content_type, chat_type, chat_id = gramscript.glance(msg)
    print(content_type, chat_type, chat_id)

    if content_type != 'successful_payment':
        sent = bot.sendInvoice(
            chat_id, "Nick's Hand Cream", "Keep a man's hand like a woman's",
            payload='a-string-identifying-related-payment-messages-tuvwxyz',
            provider_token=PAYMENT_PROVIDER_TOKEN,
            start_parameter='abc',
            currency='HKD', prices=[
                LabeledPrice(label='One Case', amount=987),
                LabeledPrice(label='Package', amount=12)],
            need_shipping_address=True, is_flexible=True)  # required for shipping query
        # 'Pay' button appears automatically

        pprint(sent)
        print(Invoice(**sent['invoice']))

    else:
        print('Successful payment RECEIVED!!!')
        pprint(msg)
        print(SuccessfulPayment(**msg['successful_payment']))


def on_shipping_query(msg):
    query_id, from_id, invoice_payload = gramscript.glance(
        msg, flavor='shipping_query')

    print('Shipping query:')
    print(query_id, from_id, invoice_payload)
    pprint(msg)
    print(ShippingQuery(**msg))

    bot.answerShippingQuery(
        query_id, True,
        shipping_options=[
            ShippingOption(id='fedex', title='FedEx', prices=[
                LabeledPrice(label='Local', amount=345),
                LabeledPrice(label='International', amount=2345)]),
            ShippingOption(id='dhl', title='DHL', prices=[
                LabeledPrice(label='Local', amount=342),
                LabeledPrice(label='International', amount=1234)])])


def on_pre_checkout_query(msg):
    query_id, from_id, invoice_payload, currency, total_amount = gramscript.glance(
        msg, flavor='pre_checkout_query', long=True)

    print('Pre-Checkout query:')
    print(query_id, from_id, invoice_payload, currency, total_amount)
    pprint(msg)
    print(PreCheckoutQuery(**msg))

    bot.answerPreCheckoutQuery(query_id, True)


TOKEN = sys.argv[1]
PAYMENT_PROVIDER_TOKEN = sys.argv[2]

bot = gramscript.Bot(TOKEN)
MessageLoop(bot, {'chat': on_chat_message,
                  'shipping_query': on_shipping_query,
                  'pre_checkout_query': on_pre_checkout_query}).run_as_thread()

while 1:
    time.sleep(10)



import collections
import warnings
import sys


class _Field(object):
    def __init__(self, name, constructor=None, default=None):
        self.name = name
        self.constructor = constructor
        self.default = default

# Function to produce namedtuple classes.


def _create_class(typename, fields):
    # extract field names
    field_names = [e.name if type(e) is _Field else e for e in fields]

    # Some dictionary keys are Python keywords and cannot be used as field names, e.g. `from`.
    # Get around by appending a '_', e.g. dict['from'] => namedtuple.from_
    keymap = [(k.rstrip('_'), k)
              for k in filter(lambda e: e in ['from_'], field_names)]

    # extract (non-simple) fields that need conversions
    conversions = [(e.name, e.constructor) for e in fields if type(
        e) is _Field and e.constructor is not None]

    # extract default values
    defaults = [e.default if type(e) is _Field else None for e in fields]

    # Create the base tuple class, with defaults.
    base = collections.namedtuple(typename, field_names)
    base.__new__.__defaults__ = tuple(defaults)

    class sub(base):
        def __new__(cls, **kwargs):
            # Map keys.
            for oldkey, newkey in keymap:
                if oldkey in kwargs:
                    kwargs[newkey] = kwargs[oldkey]
                    del kwargs[oldkey]

            # Any unexpected arguments?
            unexpected = set(kwargs.keys()) - set(super(sub, cls)._fields)

            # Remove unexpected arguments and issue warning.
            if unexpected:
                for k in unexpected:
                    del kwargs[k]

                s = ('Unexpected fields: ' + ', '.join(unexpected) + ''
                     '\nBot API seems to have added new fields to the returned data.'
                     ' This version of namedtuple is not able to capture them.'
                     '\n\nPlease upgrade gramscript by:'
                     '\n  sudo pip install gramscript --upgrade'
                     '\n\nIf you still see this message after upgrade, that means I am still working to bring the code up-to-date.'
                     ' Please try upgrade again a few days later.'
                     ' In the meantime, you can access the new fields the old-fashioned way, through the raw dictionary.')

                warnings.warn(s, UserWarning)

            # Convert non-simple values to namedtuples.
            for key, func in conversions:
                if key in kwargs:
                    if type(kwargs[key]) is dict:
                        kwargs[key] = func(**kwargs[key])
                    elif type(kwargs[key]) is list:
                        kwargs[key] = func(kwargs[key])
                    else:
                        raise RuntimeError('Can only convert dict or list')

            return super(sub, cls).__new__(cls, **kwargs)

    # https://bugs.python.org/issue24931
    # Python 3.4 bug: namedtuple subclass does not inherit __dict__ properly.
    # Fix it manually.
    if sys.version_info >= (3, 4):
        def _asdict(self):
            return collections.OrderedDict(zip(self._fields, self))
        sub._asdict = _asdict

    sub.__name__ = typename

    return sub


"""
Different treatments for incoming and outgoing namedtuples:

- Incoming ones require type declarations for certain fields for deeper parsing.
- Outgoing ones need no such declarations because users are expected to put the correct object in place.
"""

# Namedtuple class will reference other namedtuple classes. Due to circular
# dependencies, it is impossible to have all class definitions ready at
# compile time. We have to dynamically obtain class reference at runtime.
# For example, the following function acts like a constructor for `Message`
# so any class can reference the Message namedtuple even before the Message
# namedtuple is defined.


def _Message(**kwargs):
    return getattr(sys.modules[__name__], 'Message')(**kwargs)


# incoming
User = _create_class('User', [
    'id',
    'is_bot',
    'first_name',
    'last_name',
    'username',
    'language_code'
])


def UserArray(data):
    return [User(**p) for p in data]


# incoming
ChatPhoto = _create_class('ChatPhoto', [
    'small_file_id',
    'big_file_id',
])

# incoming
Chat = _create_class('Chat', [
    'id',
    'type',
    'title',
    'username',
    'first_name',
    'last_name',
    'all_members_are_administrators',
    _Field('photo', constructor=ChatPhoto),
    'description',
    'invite_link',
    _Field('pinned_message', constructor=_Message),
    'sticker_set_name',
    'can_set_sticker_set',
])

# incoming
PhotoSize = _create_class('PhotoSize', [
    'file_id',
    'width',
    'height',
    'file_size',
    'file_path',  # undocumented
])

# incoming
Audio = _create_class('Audio', [
    'file_id',
    'duration',
    'performer',
    'title',
    'mime_type',
    'file_size'
])

# incoming
Document = _create_class('Document', [
    'file_id',
    _Field('thumb', constructor=PhotoSize),
    'file_name',
    'mime_type',
    'file_size',
    'file_path',  # undocumented
])

# incoming and outgoing
MaskPosition = _create_class('MaskPosition', [
    'point',
    'x_shift',
    'y_shift',
    'scale',
])

# incoming
Sticker = _create_class('Sticker', [
    'file_id',
    'width',
    'height',
    _Field('thumb', constructor=PhotoSize),
    'emoji',
    'set_name',
    _Field('mask_position', constructor=MaskPosition),
    'file_size',
])


def StickerArray(data):
    return [Sticker(**p) for p in data]


# incoming
StickerSet = _create_class('StickerSet', [
    'name',
    'title',
    'contains_masks',
    _Field('stickers', constructor=StickerArray),
])

# incoming
Video = _create_class('Video', [
    'file_id',
    'width',
    'height',
    'duration',
    _Field('thumb', constructor=PhotoSize),
    'mime_type',
    'file_size',
    'file_path',  # undocumented
])

# incoming
Voice = _create_class('Voice', [
    'file_id',
    'duration',
    'mime_type',
    'file_size'
])

# incoming
VideoNote = _create_class('VideoNote', [
    'file_id',
    'length',
    'duration',
    _Field('thumb', constructor=PhotoSize),
    'file_size'
])

# incoming
Contact = _create_class('Contact', [
    'phone_number',
    'first_name',
    'last_name',
    'user_id'
])

# incoming
Location = _create_class('Location', [
    'longitude',
    'latitude'
])

# incoming
Venue = _create_class('Venue', [
    _Field('location', constructor=Location),
    'title',
    'address',
    'foursquare_id',
])

# incoming
File = _create_class('File', [
    'file_id',
    'file_size',
    'file_path'
])


def PhotoSizeArray(data):
    return [PhotoSize(**p) for p in data]


def PhotoSizeArrayArray(data):
    return [[PhotoSize(**p) for p in array] for array in data]


# incoming
UserProfilePhotos = _create_class('UserProfilePhotos', [
    'total_count',
    _Field('photos', constructor=PhotoSizeArrayArray)
])

# incoming
ChatMember = _create_class('ChatMember', [
    _Field('user', constructor=User),
    'status',
    'until_date',
    'can_be_edited',
    'can_change_info',
    'can_post_messages',
    'can_edit_messages',
    'can_delete_messages',
    'can_invite_users',
    'can_restrict_members',
    'can_pin_messages',
    'can_promote_members',
    'can_send_messages',
    'can_send_media_messages',
    'can_send_other_messages',
    'can_add_web_page_previews',
])


def ChatMemberArray(data):
    return [ChatMember(**p) for p in data]


# outgoing
ReplyKeyboardMarkup = _create_class('ReplyKeyboardMarkup', [
    'keyboard',
    'resize_keyboard',
    'one_time_keyboard',
    'selective',
])

# outgoing
KeyboardButton = _create_class('KeyboardButton', [
    'text',
    'request_contact',
    'request_location',
])

# outgoing
ReplyKeyboardRemove = _create_class('ReplyKeyboardRemove', [
    _Field('remove_keyboard', default=True),
    'selective',
])

# outgoing
ForceReply = _create_class('ForceReply', [
    _Field('force_reply', default=True),
    'selective',
])

# outgoing
InlineKeyboardButton = _create_class('InlineKeyboardButton', [
    'text',
    'url',
    'callback_data',
    'switch_inline_query',
    'switch_inline_query_current_chat',
    'callback_game',
    'pay',
])

# outgoing
InlineKeyboardMarkup = _create_class('InlineKeyboardMarkup', [
    'inline_keyboard',
])

# incoming
MessageEntity = _create_class('MessageEntity', [
    'type',
    'offset',
    'length',
    'url',
    _Field('user', constructor=User),
])

# incoming


def MessageEntityArray(data):
    return [MessageEntity(**p) for p in data]


# incoming
GameHighScore = _create_class('GameHighScore', [
    'position',
    _Field('user', constructor=User),
    'score',
])

# incoming
Animation = _create_class('Animation', [
    'file_id',
    _Field('thumb', constructor=PhotoSize),
    'file_name',
    'mime_type',
    'file_size',
])

# incoming
Game = _create_class('Game', [
    'title',
    'description',
    _Field('photo', constructor=PhotoSizeArray),
    'text',
    _Field('text_entities', constructor=MessageEntityArray),
    _Field('animation', constructor=Animation),
])

# incoming
Invoice = _create_class('Invoice', [
    'title',
    'description',
    'start_parameter',
    'currency',
    'total_amount',
])

# outgoing
LabeledPrice = _create_class('LabeledPrice', [
    'label',
    'amount',
])

# outgoing
ShippingOption = _create_class('ShippingOption', [
    'id',
    'title',
    'prices',
])

# incoming
ShippingAddress = _create_class('ShippingAddress', [
    'country_code',
    'state',
    'city',
    'street_line1',
    'street_line2',
    'post_code',
])

# incoming
OrderInfo = _create_class('OrderInfo', [
    'name',
    'phone_number',
    'email',
    _Field('shipping_address', constructor=ShippingAddress),
])

# incoming
ShippingQuery = _create_class('ShippingQuery', [
    'id',
    _Field('from_', constructor=User),
    'invoice_payload',
    _Field('shipping_address', constructor=ShippingAddress),
])

# incoming
PreCheckoutQuery = _create_class('PreCheckoutQuery', [
    'id',
    _Field('from_', constructor=User),
    'currency',
    'total_amount',
    'invoice_payload',
    'shipping_option_id',
    _Field('order_info', constructor=OrderInfo),
])

# incoming
SuccessfulPayment = _create_class('SuccessfulPayment', [
    'currency',
    'total_amount',
    'invoice_payload',
    'shipping_option_id',
    _Field('order_info', constructor=OrderInfo),
    'telegram_payment_charge_id',
    'provider_payment_charge_id',
])

# incoming
Message = _create_class('Message', [
    'message_id',
    _Field('from_', constructor=User),
    'date',
    _Field('chat', constructor=Chat),
    _Field('forward_from', constructor=User),
    _Field('forward_from_chat', constructor=Chat),
    'forward_from_message_id',
    'forward_signature',
    'forward_date',
    _Field('reply_to_message', constructor=_Message),
    'edit_date',
    'author_signature',
    'text',
    _Field('entities', constructor=MessageEntityArray),
    _Field('caption_entities', constructor=MessageEntityArray),
    _Field('audio', constructor=Audio),
    _Field('document', constructor=Document),
    _Field('game', constructor=Game),
    _Field('photo', constructor=PhotoSizeArray),
    _Field('sticker', constructor=Sticker),
    _Field('video', constructor=Video),
    _Field('voice', constructor=Voice),
    _Field('video_note', constructor=VideoNote),
    _Field('new_chat_members', constructor=UserArray),
    'caption',
    _Field('contact', constructor=Contact),
    _Field('location', constructor=Location),
    _Field('venue', constructor=Venue),
    _Field('new_chat_member', constructor=User),
    _Field('left_chat_member', constructor=User),
    'new_chat_title',
    _Field('new_chat_photo', constructor=PhotoSizeArray),
    'delete_chat_photo',
    'group_chat_created',
    'supergroup_chat_created',
    'channel_chat_created',
    'migrate_to_chat_id',
    'migrate_from_chat_id',
    _Field('pinned_message', constructor=_Message),
    _Field('invoice', constructor=Invoice),
    _Field('successful_payment', constructor=SuccessfulPayment),
    'connected_website',
])

# incoming
InlineQuery = _create_class('InlineQuery', [
    'id',
    _Field('from_', constructor=User),
    _Field('location', constructor=Location),
    'query',
    'offset',
])

# incoming
ChosenInlineResult = _create_class('ChosenInlineResult', [
    'result_id',
    _Field('from_', constructor=User),
    _Field('location', constructor=Location),
    'inline_message_id',
    'query',
])

# incoming
CallbackQuery = _create_class('CallbackQuery', [
    'id',
    _Field('from_', constructor=User),
    _Field('message', constructor=Message),
    'inline_message_id',
    'chat_instance',
    'data',
    'game_short_name',
])

# incoming
Update = _create_class('Update', [
    'update_id',
    _Field('message', constructor=Message),
    _Field('edited_message', constructor=Message),
    _Field('channel_post', constructor=Message),
    _Field('edited_channel_post', constructor=Message),
    _Field('inline_query', constructor=InlineQuery),
    _Field('chosen_inline_result', constructor=ChosenInlineResult),
    _Field('callback_query', constructor=CallbackQuery),
])

# incoming


def UpdateArray(data):
    return [Update(**u) for u in data]


# incoming
WebhookInfo = _create_class('WebhookInfo', [
    'url',
    'has_custom_certificate',
    'pending_update_count',
    'last_error_date',
    'last_error_message',
])

# outgoing
InputTextMessageContent = _create_class('InputTextMessageContent', [
    'message_text',
    'parse_mode',
    'disable_web_page_preview',
])

# outgoing
InputLocationMessageContent = _create_class('InputLocationMessageContent', [
    'latitude',
    'longitude',
    'live_period',
])

# outgoing
InputVenueMessageContent = _create_class('InputVenueMessageContent', [
    'latitude',
    'longitude',
    'title',
    'address',
    'foursquare_id',
])

# outgoing
InputContactMessageContent = _create_class('InputContactMessageContent', [
    'phone_number',
    'first_name',
    'last_name',
])

# outgoing
InlineQueryResultArticle = _create_class('InlineQueryResultArticle', [
    _Field('type', default='article'),
    'id',
    'title',
    'input_message_content',
    'reply_markup',
    'url',
    'hide_url',
    'description',
    'thumb_url',
    'thumb_width',
    'thumb_height',
])

# outgoing
InlineQueryResultPhoto = _create_class('InlineQueryResultPhoto', [
    _Field('type', default='photo'),
    'id',
    'photo_url',
    'thumb_url',
    'photo_width',
    'photo_height',
    'title',
    'description',
    'caption',
    'parse_mode',
    'reply_markup',
    'input_message_content',
])

# outgoing
InlineQueryResultGif = _create_class('InlineQueryResultGif', [
    _Field('type', default='gif'),
    'id',
    'gif_url',
    'gif_width',
    'gif_height',
    'gif_duration',
    'thumb_url',
    'title',
    'caption',
    'parse_mode',
    'reply_markup',
    'input_message_content',
])

# outgoing
InlineQueryResultMpeg4Gif = _create_class('InlineQueryResultMpeg4Gif', [
    _Field('type', default='mpeg4_gif'),
    'id',
    'mpeg4_url',
    'mpeg4_width',
    'mpeg4_height',
    'mpeg4_duration',
    'thumb_url',
    'title',
    'caption',
    'parse_mode',
    'reply_markup',
    'input_message_content',
])

# outgoing
InlineQueryResultVideo = _create_class('InlineQueryResultVideo', [
    _Field('type', default='video'),
    'id',
    'video_url',
    'mime_type',
    'thumb_url',
    'title',
    'caption',
    'parse_mode',
    'video_width',
    'video_height',
    'video_duration',
    'description',
    'reply_markup',
    'input_message_content',
])

# outgoing
InlineQueryResultAudio = _create_class('InlineQueryResultAudio', [
    _Field('type', default='audio'),
    'id',
    'audio_url',
    'title',
    'caption',
    'parse_mode',
    'performer',
    'audio_duration',
    'reply_markup',
    'input_message_content',
])

# outgoing
InlineQueryResultVoice = _create_class('InlineQueryResultVoice', [
    _Field('type', default='voice'),
    'id',
    'voice_url',
    'title',
    'caption',
    'parse_mode',
    'voice_duration',
    'reply_markup',
    'input_message_content',
])

# outgoing
InlineQueryResultDocument = _create_class('InlineQueryResultDocument', [
    _Field('type', default='document'),
    'id',
    'title',
    'caption',
    'parse_mode',
    'document_url',
    'mime_type',
    'description',
    'reply_markup',
    'input_message_content',
    'thumb_url',
    'thumb_width',
    'thumb_height',
])

# outgoing
InlineQueryResultLocation = _create_class('InlineQueryResultLocation', [
    _Field('type', default='location'),
    'id',
    'latitude',
    'longitude',
    'title',
    'live_period',
    'reply_markup',
    'input_message_content',
    'thumb_url',
    'thumb_width',
    'thumb_height',
])

# outgoing
InlineQueryResultVenue = _create_class('InlineQueryResultVenue', [
    _Field('type', default='venue'),
    'id',
    'latitude',
    'longitude',
    'title',
    'address',
    'foursquare_id',
    'reply_markup',
    'input_message_content',
    'thumb_url',
    'thumb_width',
    'thumb_height',
])

# outgoing
InlineQueryResultContact = _create_class('InlineQueryResultContact', [
    _Field('type', default='contact'),
    'id',
    'phone_number',
    'first_name',
    'last_name',
    'reply_markup',
    'input_message_content',
    'thumb_url',
    'thumb_width',
    'thumb_height',
])

# outgoing
InlineQueryResultGame = _create_class('InlineQueryResultGame', [
    _Field('type', default='game'),
    'id',
    'game_short_name',
    'reply_markup',
])

# outgoing
InlineQueryResultCachedPhoto = _create_class('InlineQueryResultCachedPhoto', [
    _Field('type', default='photo'),
    'id',
    'photo_file_id',
    'title',
    'description',
    'caption',
    'parse_mode',
    'reply_markup',
    'input_message_content',
])

# outgoing
InlineQueryResultCachedGif = _create_class('InlineQueryResultCachedGif', [
    _Field('type', default='gif'),
    'id',
    'gif_file_id',
    'title',
    'caption',
    'parse_mode',
    'reply_markup',
    'input_message_content',
])

# outgoing
InlineQueryResultCachedMpeg4Gif = _create_class('InlineQueryResultCachedMpeg4Gif', [
    _Field('type', default='mpeg4_gif'),
    'id',
    'mpeg4_file_id',
    'title',
    'caption',
    'parse_mode',
    'reply_markup',
    'input_message_content',
])

# outgoing
InlineQueryResultCachedSticker = _create_class('InlineQueryResultCachedSticker', [
    _Field('type', default='sticker'),
    'id',
    'sticker_file_id',
    'reply_markup',
    'input_message_content',
])

# outgoing
InlineQueryResultCachedDocument = _create_class('InlineQueryResultCachedDocument', [
    _Field('type', default='document'),
    'id',
    'title',
    'document_file_id',
    'description',
    'caption',
    'parse_mode',
    'reply_markup',
    'input_message_content',
])

# outgoing
InlineQueryResultCachedVideo = _create_class('InlineQueryResultCachedVideo', [
    _Field('type', default='video'),
    'id',
    'video_file_id',
    'title',
    'description',
    'caption',
    'parse_mode',
    'reply_markup',
    'input_message_content',
])

# outgoing
InlineQueryResultCachedVoice = _create_class('InlineQueryResultCachedVoice', [
    _Field('type', default='voice'),
    'id',
    'voice_file_id',
    'title',
    'caption',
    'parse_mode',
    'reply_markup',
    'input_message_content',
])

# outgoing
InlineQueryResultCachedAudio = _create_class('InlineQueryResultCachedAudio', [
    _Field('type', default='audio'),
    'id',
    'audio_file_id',
    'caption',
    'parse_mode',
    'reply_markup',
    'input_message_content',
])

# outgoing
InputMediaPhoto = _create_class('InputMediaPhoto', [
    _Field('type', default='photo'),
    'media',
    'caption',
    'parse_mode',
])

# outgoing
InputMediaVideo = _create_class('InputMediaVideo', [
    _Field('type', default='video'),
    'media',
    'caption',
    'parse_mode',
    'width',
    'height',
    'duration',
    'supports_streaming',
])

# incoming
ResponseParameters = _create_class('ResponseParameters', [
    'migrate_to_chat_id',
    'retry_after',
])



import filecmp
import io
import sys
import urllib3
import logging
import json
import re
import os

from . import exception

# Suppress InsecurePlatformWarning
urllib3.disable_warnings()


_default_pool_params = dict(num_pools=3, maxsize=10, retries=3, timeout=30)
_onetime_pool_params = dict(num_pools=1, maxsize=1, retries=3, timeout=30)

_pools = {
    'default': urllib3.PoolManager(**_default_pool_params),
}

_onetime_pool_spec = (urllib3.PoolManager, _onetime_pool_params)
PY_3 = sys.version_info.major >= 3
_string_type = str if PY_3 else basestring
_file_type = io.IOBase if PY_3 else filecmp


def _isstring(s):
    return isinstance(s, _string_type)


def _isfile(f):
    return isinstance(f, _file_type)


def set_proxy(url, basic_auth=None):
    """
    Access Bot API through a proxy.

    :param url: proxy URL
    :param basic_auth: 2-tuple ``('username', 'password')``
    """
    global _pools, _onetime_pool_spec
    if not url:
        _pools['default'] = urllib3.PoolManager(**_default_pool_params)
        _onetime_pool_spec = (urllib3.PoolManager, _onetime_pool_params)
    elif basic_auth:
        h = urllib3.make_headers(proxy_basic_auth=':'.join(basic_auth))
        _pools['default'] = urllib3.ProxyManager(
            url, proxy_headers=h, **_default_pool_params)
        _onetime_pool_spec = (urllib3.ProxyManager, dict(
            proxy_url=url, proxy_headers=h, **_onetime_pool_params))
    else:
        _pools['default'] = urllib3.ProxyManager(url, **_default_pool_params)
        _onetime_pool_spec = (urllib3.ProxyManager, dict(
            proxy_url=url, **_onetime_pool_params))


def _create_onetime_pool():
    cls, kw = _onetime_pool_spec
    return cls(**kw)


def _methodurl(req, **user_kw):
    token, method, params, files = req
    return f'https://api.telegram.org/bot{token}/{method}'


def _which_pool(req, **user_kw):
    token, method, params, files = req
    return None if files else 'default'


def _guess_filename(obj):
    name = getattr(obj, 'name', None)
    if name and _isstring(name) and name[0] != '<' and name[-1] != '>':
        return os.path.basename(name)


def _filetuple(key, f):
    if not isinstance(f, tuple):
        return (_guess_filename(f) or key, f.read())
    elif len(f) == 1:
        return (_guess_filename(f[0]) or key, f[0].read())
    elif len(f) == 2:
        return (f[0], f[1].read())
    elif len(f) == 3:
        return (f[0], f[1].read(), f[2])
    else:
        raise ValueError()


PY_3 = sys.version_info.major >= 3


def _fix_type(v):
    return str(v) if isinstance(v, float if PY_3 else (long, float)) else v


def _compose_fields(req, **user_kw):
    token, method, params, files = req
    fields = {k: _fix_type(v) for k, v in params.items()
              } if params is not None else {}

    if files:
        fields |= {k: _filetuple(k, v) for k, v in files.items()}
    return fields


def _default_timeout(req, **user_kw):
    name = _which_pool(req, **user_kw)
    if name is None:
        return _onetime_pool_spec[1]['timeout']
    else:
        return _pools[name].connection_pool_kw['timeout']


def _compose_kwargs(req, **user_kw):
    token, method, params, files = req
    kw = {}
    if not params and not files:
        kw['encode_multipart'] = False
    if method == 'getUpdates' and params and 'timeout' in params:
        kw['timeout'] = params['timeout'] + _default_timeout(req, **user_kw)
    elif files:
        kw['timeout'] = None
    kw |= user_kw
    return kw


def _transform(req, **user_kw):
    kwargs = _compose_kwargs(req, **user_kw)
    fields = _compose_fields(req, **user_kw)
    url = _methodurl(req, **user_kw)
    name = _which_pool(req, **user_kw)
    pool = _create_onetime_pool() if name is None else _pools[name]
    return pool.request_encode_body, ('POST', url, fields), kwargs


def _parse(response):
    try:
        text = response.data.decode('utf-8')
        data = json.loads(text)
    except ValueError as e:
        raise exception.BadHTTPResponse(response.status, text, response) from e
    if data['ok']:
        return data['result']
    description, error_code = data['description'], data['error_code']
    for e in exception.TelegramError.__subclasses__():
        n = len(e.DESCRIPTION_PATTERNS)
        if any(map(re.search, e.DESCRIPTION_PATTERNS, n * [description], n * [re.IGNORECASE])):
            raise e(description, error_code, data)
    raise exception.TelegramError(description, error_code, data)


def request(req, **user_kw):
    fn, args, kwargs = _transform(req, **user_kw)
    r = fn(*args, **kwargs)  # `fn` must be thread-safe
    return _parse(r)


def _fileurl(req):
    token, path = req
    return f'https://api.telegram.org/file/bot{token}/{path}'


def download(req, **user_kw):
    pool = _create_onetime_pool()
    return pool.request('GET', _fileurl(req), **user_kw)



# coding=utf8

import time
import threading
import pprint
import sys
import traceback
import gramscript
import gramscript.namedtuple

"""
This script tests:
- receiving all types of messages, by asking user to produce each

Run it by:
$ python3.X test3.py <token> <user_id>

It will assume the bot identified by <token>, and only communicate with the user identified by <user_id>.

If you don't know your user id, run:
$ python test.py <token> 0

And send it a message anyway. It will print out your user id as an unauthorized user.
Ctrl-C to kill it, then run the proper command again.
"""


def equivalent(data, nt):
    if type(data) is dict:
        keys = list(data.keys())

        # number of dictionary keys == number of non-None values in namedtuple?
        if len(keys) != len([f for f in nt._fields if getattr(nt, f) is not None]):
            return False

        # map `from` to `from_`
        fields = [f'{k}_' if k in ['from'] else k for k in keys]

        return all(map(equivalent, [data[k] for k in keys], [getattr(nt, f) for f in fields]))
    elif type(data) is list:
        return all(map(equivalent, data, nt))
    else:
        return data == nt


def examine(result, type):
    try:
        print(f'Examining {type} ......')
        nt = type(**result)
        assert equivalent(
            result, nt), 'Not equivalent:::::::::::::::\n%s\n::::::::::::::::\n%s' % (result, nt)

        pprint.pprint(result)
        pprint.pprint(nt)
        print()
    except AssertionError:
        traceback.print_exc()
        answer = input('Do you want to continue? [y] ')
        if answer != 'y':
            exit(1)


expected_content_type = None
content_type_iterator = iter([
    'text', 'voice', 'sticker', 'photo', 'audio', 'document', 'video', 'contact', 'location',
    'new_chat_member',  'new_chat_title', 'new_chat_photo',  'delete_chat_photo', 'left_chat_member'
])


def see_every_content_types(msg):
    global expected_content_type, content_type_iterator
    content_type, chat_type, chat_id = gramscript.glance(msg)
    from_id = msg['from']['id']
    if chat_id != USER_ID and from_id != USER_ID:
        print('Unauthorized user:', chat_id, from_id)
        return
    examine(msg, gramscript.namedtuple.Message)
    try:
        if content_type == expected_content_type:
            expected_content_type = next(content_type_iterator)
            bot.sendMessage(
                chat_id, f'Please give me a {expected_content_type}.')
        else:
            bot.sendMessage(
                chat_id, f'It is not a {expected_content_type}. Please give me a {expected_content_type}, please.')

    except StopIteration:
        bot.sendMessage(from_id, 'Thank you. I am done.')


TOKEN = sys.argv[1]
USER_ID = int(sys.argv[2])

# gramscript.api.set_proxy('http://192.168.0.103:3128')

bot = gramscript.Bot(TOKEN)

expected_content_type = content_type_iterator.__next__()
bot.sendMessage(USER_ID, f'Please give me a {expected_content_type}.')
bot.message_loop(see_every_content_types, run_forever=True)
